char *title = "Losing Chess three-piece database interrogation program";
char *version = "1.0";
char *authorship = "JDB, 20 Jan 1998";

/* Please alter these opening lines if you produce your own version */

/*	Command is

		LOSING3 log_file

	The log file receives a copy of the screen output

	If "log_file" is omitted, LOSING3.LOG is assumed

	The program uses database files LOSING3x.TXT where "x" is K, N, B, R, Q

*/

/* In Turbo C, compile using the COMPACT memory model */

/* Warning: tabs are set at 3 characters */

/* Standard Turbo C header files (may include some not needed here) */

#include <alloc.h>
#include <conio.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys\stat.h>

/* general constants */

#define TRUE 1
#define FALSE 0

/* constants defining the men and their powers */

#define LETTER "KNBRQ"
#define ROW_INC 10

enum men
 {
	KING, KNIGHT, BISHOP, ROOK, QUEEN, EMPTY, NOT_ON_BOARD
 };

enum move_type
 {
	LEAPER, RIDER
 };

int king_move[9] =
 {
	- ROW_INC - 1, - ROW_INC, - ROW_INC + 1, - 1,
	1, ROW_INC - 1, ROW_INC, ROW_INC + 1,
	0
 };

int knight_move[9] =
 {
	- 2 * ROW_INC - 1, - 2 * ROW_INC + 1,
	- ROW_INC - 2, - ROW_INC + 2,
	ROW_INC - 2, ROW_INC + 2,
	2 * ROW_INC - 1, 2 * ROW_INC + 1,
	0
 };

int bishop_move[5] =
 {
	- ROW_INC - 1, - ROW_INC + 1,
	ROW_INC - 1, ROW_INC + 1,
	0
 };

int rook_move[5] =
 {
	- ROW_INC, - 1, 1, ROW_INC,
	0
 };

int *move_power[5] =
 {
	king_move, knight_move, bishop_move, rook_move, king_move
 };

int move_type[5] =
 {
	LEAPER, LEAPER, RIDER, RIDER, RIDER
 };

enum men_to_move
 {
	W_MAN, B1_MAN, B2_MAN
 };

/* constants related to input processing */

#define CONSOLE_INPUT_LIMIT 20
#define ARTIFICIAL_INPUT_TERMINATOR '*'

enum console_input_states
 {
	READING_FIRST_CHARACTER, READING_MAN, READING_COL, READING_ROW,
	READING_SIDE_TO_MOVE, READING_INTEGER,
	SKIPPING_TO_END
 };

enum console_actions
 {
	SET_POSITION, OBEY_QUERIES, GIVE_HELP, CHOOSE_MOVE, RETRACT_MOVES,
	REPEAT_POSITION, LIST_MOVES_PLAYED, REPORT_STATISTICS,
	DO_NOTHING, REPORT_ERROR, EXIT
 };

enum order_of_black_men
 {
	NORMAL, INVERTED
 };

enum database_input_states
 {
	FINDING_SECTION, READING_PLAYERS,
	READING_RESULTS, READING_CHECK_SUM
 };

/* function prototypes */

void analyse_black_position(void);
void analyse_complete_database(void);
void analyse_white_position(void);
int board_square(int row, int column);
void calculate_moves(int man, int row, int col, int man_number);
void clear_board(void);
void clear_target_board(void);
int colour_test(int b1r, int b1c, int b2r, int b2c);
void encode_address(int *slice, int *address,
	int b1_row, int b1_col, int b2_row, int b2_col, int w_row, int w_col);
int find_char(char *string, int c);
int good_address(int b1r, int b1c, int b2r, int b2c, int wr, int wc);
void list_and_print_moves(void);
void list_moves_played(void);
void list_positions(int items, int limit, unsigned char item[][3],
	int m1, int m2, char *title);
int print_and_echo_result(int datum);
void print_help_screen(void);
void print_position_and_moves(void);
void read_and_obey_console_command(void);
int read_database(void);
void report_run_statistics(void);
void set_board(void);
void set_position(char *string,
	int b1r, int b1c, int b2r, int b2c, int wr, int wc);

/* global declarations and associated constants */

/* memory for in-store database */
/* "huge" is Turbo C jargon for a 32-bit pointer */

#define MEMORY_PER_SLICE (long) 4096
#define TOTAL_MEMORY (long) (20 * MEMORY_PER_SLICE)

void huge *start_of_memory;
unsigned char huge *wtm_data[10], huge *btm_data[10], huge *data_ptr;
long total_memory;
unsigned char letter_map[256], slice_map[8][8], row_map[10], col_map[10];

/* database files and associated variables */

FILE *input_file[5], *log_file;
char *input_filename[5], *log_filename;
int input_handle[5], log_handle;
int brackets_read[5];
int section_map[5][5];

/* console input */

char console_input_buffer[CONSOLE_INPUT_LIMIT + 4];
int input_count, quit, position_set;
char opening_position[14], current_position[14];
int w_man, w_row, w_col, b1_man, b1_row, b1_col, b2_man, b2_row, b2_col, wtm,
	last_w_man, last_b1_man, last_b2_man, order_of_b_men, w_queried, b_queried;

/* working boards */

#define MAX_MOVE_INC (2 * ROW_INC + 1)
#define BOARD_START MAX_MOVE_INC
#define BOARD_END (BOARD_START + 7 * ROW_INC + 7)
#define BOARD_ARRAY_SIZE (BOARD_END + MAX_MOVE_INC + 1)

int board[BOARD_ARRAY_SIZE], target_board[BOARD_ARRAY_SIZE],
	man_start, *increment_ptr, man_type;

/* lists of moves available and played */

#define MOVES_AVAILABLE_LIMIT 54  /* two queens */
#define MOVES_PLAYED_LIMIT 100    /* more than enough for any sensible line! */

int move[MOVES_AVAILABLE_LIMIT][8], moves,
	best_result, best_move, unique_best_move;
int move_played[MOVES_PLAYED_LIMIT + 1][13], moves_played;

/* individual database summaries */

#define DRAW_LIMIT 600
#define LONGEST_WIN_LIMIT 50
#define LONGEST_LOSS_LIMIT 50
#define LOSS_WIN_LIMIT 100
#define LOSS_DRAW_LIMIT 400
#define DRAW_WIN_LIMIT 400

unsigned char draw[DRAW_LIMIT][3], longest_win[LONGEST_WIN_LIMIT][3],
	longest_loss[LONGEST_LOSS_LIMIT][3], loss_win[LOSS_WIN_LIMIT][3],
	loss_draw[LOSS_DRAW_LIMIT][3], draw_win[DRAW_WIN_LIMIT][3];
int draws, longest_win_so_far, longest_wins, longest_loss_so_far,
	longest_losses, loss_wins, loss_draws, draw_wins;

/* overall statistics */

#define ANALYSIS_LIMIT 150
unsigned char men[ANALYSIS_LIMIT][4];
int analysis_data[ANALYSIS_LIMIT][13], analyses;

/* program starts here */

main (int number_of_arguments, char *argument[])

 {
	char buffer[81], new[7];
	int count, c, i;
	long position_count, row, col;
	int ok_to_continue;

	printf("\n");
	printf("%s\n", title);
	printf("Version %s (%s)\n", version, authorship);

	log_filename = (number_of_arguments > 1) ? argument[1] : "LOSING3.LOG";
	input_filename[0] = "LOSING3K.TXT";
	input_filename[1] = "LOSING3N.TXT";
	input_filename[2] = "LOSING3B.TXT";
	input_filename[3] = "LOSING3R.TXT";
	input_filename[4] = "LOSING3Q.TXT";

	remove(log_filename);
	log_handle = open(log_filename, O_CREAT, S_IREAD | S_IWRITE);
	log_file = fdopen(log_handle, "wt");
	if (log_file == NULL)
	 {
		printf("\n");
		printf("Unable to open log file %s\n", log_filename);
		ok_to_continue = FALSE;
	 }
	else
	 {
		printf("\n");
		printf("A copy of the screen output is being sent to file %s\n",
			log_filename);
		fprintf(log_file, "Written by %s, version %s\n", title, version);
		ok_to_continue = TRUE;
	 }

/* next section (memory allocation and pointer initialization) may be
	Turbo C dependent */

	if (ok_to_continue)
	 {
		total_memory = (long) TOTAL_MEMORY;
		printf("\n");
		printf("   Memory required:   %8lu\n", total_memory);
		printf("   Memory available:  %8lu\n", farcoreleft());
		start_of_memory = farmalloc(total_memory);
		if (start_of_memory == NULL)
		 {
			printf("   Unable to allocate memory\n");
			ok_to_continue = FALSE;
		 }
		else
		 {
			printf("   Memory allocated\n");

			data_ptr = start_of_memory;
			for (count= 0; count < 10; count++)
			 {
				wtm_data[count] = data_ptr;
				data_ptr += MEMORY_PER_SLICE;
			 }
			for (count= 0; count < 10; count++)
			 {
				btm_data[count] = data_ptr;
				data_ptr += MEMORY_PER_SLICE;
			 }

/* insertion for test purposes ...

			printf("\n");
			printf("Memory pointers:\n");
			printf("\n");
			for (count= 0; count < 10; count++)
				printf("%10p  %10p\n", wtm_data[count], btm_data[count]);

... end of test insertion */

		 }
	 }

	if (ok_to_continue)
	 {
		printf("\n");
		for (count = 0; count < 5; count++)
		 {
			input_handle[count] = open(input_filename[count], O_RDONLY);
			input_file[count] = fdopen(input_handle[count], "rt");
			if (input_file[count] == NULL)
			 {
				printf("   Unable to open database file %s\n",
					input_filename[count]);
				brackets_read[count] = - 1;
			 }
			else
			 {
				printf("   Database file %s opened\n", input_filename[count]);
				brackets_read[count] = 0;
			 }
		 }

		count = 0;
		for (b1_man = 0; b1_man < 5; b1_man++)
		 {
			for (b2_man = 0; b2_man <= b1_man; b2_man++)
			 {
				count++;
				section_map[b1_man][b2_man] = section_map[b2_man][b1_man] = count;
			 }
		}

		for (count = 0; count < 256; count++)
			letter_map[count] = -1;
		for (count = 0; count < 5; count ++)
			letter_map[(LETTER[count])] = count;

		count = 0;
		for (row = 7; row >= 4; row--)
		 {
			for (col = 7 - row; col < 4; col++)
			 {
				slice_map[row][col] = count;
				slice_map[7 - row][col] = count;
				slice_map[row][7 - col] = count;
				slice_map[7 - row][7 - col] = count;
				slice_map[col][row] = count;
				slice_map[col][7 - row] = count;
				slice_map[7 - col][row] = count;
				slice_map[7 - col][7 - row] = count;
				row_map[count] = row;
				col_map[count] = col;
				count++;
			 }
		 }

		analyses = 0;
		last_w_man = - 1;
		position_set = FALSE;
		moves = 0;
   	quit = FALSE;

		do
		 {
			read_and_obey_console_command();
		 } while (!quit);

		for (count = 0; count < 5; count++)
		 {
			if (brackets_read[count] >= 0)
				fclose(input_file[count]);
		 }
		fclose(log_file);

		printf("\n");
		printf("A copy of the screen output has been saved in file %s\n",
			log_filename);
		if (number_of_arguments <= 1)
		 {
			printf("WARNING: the next use of LOSING3 without a parameter will overwrite this file\n");
			printf("Rename it now if it contains anything important\n");
		 }
	 }
 }

/* functions follow in alphabetical order */

void analyse_black_position()
 {
	int pass, slice, position, datum, value;
	char code;

	set_position(opening_position,
		b1_row, b1_col, b2_row, b2_col, 0, 0);

	printf("\n");
	printf("   Analysis of %c v %s\n", opening_position[0],
		opening_position + 6);
	printf("\n");
	printf("              White to move                      Black to move\n");
	printf("\n");
	fprintf(log_file, "\n");
	fprintf(log_file, "Analysis of %c v %s\n", opening_position[0],
		opening_position + 6);
	fprintf(log_file, "\n");
	fprintf(log_file, "              White to move                      Black to move\n");
	fprintf(log_file, "\n");
	for (w_row = 7; w_row >= 0; w_row--)
	 {
		for (pass = 0; pass < 2; pass++)
		 {
			printf("   ");
			fprintf(log_file, "   ");
			for (w_col = 0; w_col < 8; w_col++)
			 {
				encode_address(&slice, &position,
					b1_row, b1_col, b2_row, b2_col, w_row, w_col);
				datum = (pass == 0) ? wtm_data[slice][position]
										  : btm_data[slice][position];
				if ( (datum >= 'A') && (datum <= 'Z') )
					value = - (datum - 'A' + 1);
				else if ( (datum == '~') || (datum == '|') )
					value = 0;
				else if (datum == '\\')
					value = 1;
				else if ( (datum >= 'a') && (datum <= 'z') )
					value = (datum - 'a' + 1);
				else
					value = - 99;
				if (value == - 99)
				 {
					if ( (w_row == b1_row) && (w_col == b1_col) )
						code = LETTER[b1_man];
					else if ( (w_row == b2_row) && (w_col == b2_col) )
						code = LETTER[b2_man];
					else
						code = '*';
					printf("  b%c", code);
					fprintf(log_file, "  b%c", code);
				 }
				else if (value == - 1)
				 {
					printf("   .");
					fprintf(log_file, "   .");
				 }
				else if (value == 0)
				 {
					printf("   =");
					fprintf(log_file, "   =");
				 }
				else
				 {
					printf(" %3d", value);
					fprintf(log_file, " %3d", value);
				 }
			 }
		 }
		printf("\n");
		fprintf(log_file, "\n");
	 }
 }

void analyse_complete_database()
 {
	int pass, count, passes, wr, wc, b1r, b1c, b2r, b2c, m1, m2,
		sum[9], grand_sum[9], pass_sum[2][9], counts[2][54], title_needed,
		max_win, max_win_row, max_win_col,
		max_loss, max_loss_row, max_loss_col,
		white_sum[10][9], white_slice,
		slice, position, datum, value, otm_datum, length,
		blank_line_needed;

	printf("\n");
	printf("   Analysing...");

	for (pass = 0; pass < 2; pass++)
	 {
		for (count = 0; count < 9; count++)
			pass_sum[pass][count] = 0;
		for (count = 0; count < 54; count++)
			counts[pass][count] = 0;
	 }
	for (white_slice = 0; white_slice < 10; white_slice++)
	 {
		for (count = 0; count < 9; count++)
			white_sum[white_slice][count] = 0;
	 }
	draws = loss_wins = loss_draws = draw_wins = longest_win_so_far
		= longest_wins = longest_loss_so_far = longest_losses = 0;

	if (order_of_b_men == NORMAL)
	 {
		m1 = b1_man;
		m2 = b2_man;
	 }
	else
	 {
		m1 = b2_man;
		m2 = b1_man;
	 }
	if ( ( (LETTER[b1_man] == 'B') && (LETTER[b2_man] == 'B') )
	  || ( (LETTER[b1_man] == 'B') && (LETTER[b2_man] == 'N') )
	  || ( (LETTER[b1_man] == 'N') && (LETTER[b2_man] == 'B') ) )
		passes = 2;
	else
		passes = 1;

	for (pass = 0; pass < passes; pass++)
	 {
		fprintf(log_file, "\n");
		fprintf(log_file, "Analysis of %c v %c+%c,", LETTER[w_man],
			LETTER[m1], LETTER[m2]);
		if (passes == 2)
			fprintf(log_file, " %slike squares,", (pass == 0) ? "un" : "");
		fprintf(log_file, " %s to move\n", (wtm) ? "White" : "Black");
		fprintf(log_file, "\n");
		fprintf(log_file, "Black       --- Wins ---       Draws      -- Losses --   Total      Max     Max\n");
		fprintf(log_file, "pieces      1     2    3+     1    2+     1     2    3+             win    loss\n");


		for (b1r = 7; b1r >= 4; b1r--)
		 {
			for (b1c = 7 - b1r; b1c < 4; b1c++)
			 {
				for (b2r = 7; b2r >= 0; b2r--)
				 {
					blank_line_needed = TRUE;
					for (b2c = 0; b2c < 8; b2c++)
					 {
						for (count = 0; count < 9; count++)
							sum[count] = 0;
						max_win = max_loss = 0;
						for (wr = 7; wr >= 0; wr--)
						 {
							for (wc = 0; wc < 8; wc++)
							 {
								if ( (colour_test(b1r, b1c, b2r, b2c) == pass)
								  && (good_address(b1r, b1c, b2r, b2c, wr, wc)) )
								 {
									if (order_of_b_men == NORMAL)
										encode_address(&slice, &position,
											b1r, b1c, b2r, b2c, wr, wc);
									else
										encode_address(&slice, &position,
											b2r, b2c, b1r, b1c, wr, wc);
									if (wtm)
									 {
										datum = wtm_data[slice][position];
										otm_datum = btm_data[slice][position];
									 }
									else
									 {
										datum = btm_data[slice][position];
										otm_datum = wtm_data[slice][position];
									 }
									white_slice = slice_map[wr][wc];
									if ( (datum == '\\') || (datum == 'a') )
									 {
										sum[0]++;
										white_sum[white_slice][0]++;
										counts[pass][0]++;
									 }
									else if (datum == 'b')
									 {
										sum[1]++;
										white_sum[white_slice][1]++;
										counts[pass][1]++;
									 }
									else if ( (datum >= 'c') && (datum <= 'z') )
									 {
										sum[2]++;
										white_sum[white_slice][2]++;
										length = datum - 'a' + 1;
										counts[pass][length - 1]++;
										if (length > max_win)
										 {
											max_win = length;
											max_win_row = wr;
											max_win_col = wc;
											if (length > longest_win_so_far)
											 {
												longest_win_so_far = length;
												longest_wins = 0;
											 }
										 }
										if (length == longest_win_so_far)
										 {
											if (longest_wins < LONGEST_WIN_LIMIT)
											 {
												longest_win[longest_wins][0] = b1r << 3 | b1c;
												longest_win[longest_wins][1] = b2r << 3 | b2c;
												longest_win[longest_wins][2] = wr << 3 | wc;
											 }
											longest_wins++;
										 }
									 }
									else if (datum == '|')
									 {
										sum[3]++;
										white_sum[white_slice][3]++;
										counts[pass][26]++;
									 }
									else if (datum == '~')
									 {
										sum[4]++;
										white_sum[white_slice][4]++;
										counts[pass][26]++;
										if (draws < DRAW_LIMIT)
										 {
											draw[draws][0] = b1r << 3 | b1c;
											draw[draws][1] = b2r << 3 | b2c;
											draw[draws][2] = wr << 3 | wc;
										 }
										draws++;
									 }
									else if (datum == 'A')
									 {
										sum[5]++;
										white_sum[white_slice][5]++;
										counts[pass][52]++;
									 }
									else if (datum == 'B')
									 {
										sum[6]++;
										white_sum[white_slice][6]++;
										counts[pass][51]++;
									 }
									else if ( (datum >= 'C') && (datum <= 'Z') )
									 {
										sum[7]++;
										white_sum[white_slice][7]++;
										length = datum - 'A' + 1;
										counts[pass][53 - length]++;
										if (length > max_loss)
										 {
											max_loss = length;
											max_loss_row = wr;
											max_loss_col = wc;
											if (length > longest_loss_so_far)
											 {
												longest_loss_so_far = length;
												longest_losses = 0;
											 }
										 }
										if (length == longest_loss_so_far)
										 {
											if (longest_losses < LONGEST_LOSS_LIMIT)
											 {
												longest_loss[longest_losses][0] = b1r << 3 | b1c;
												longest_loss[longest_losses][1] = b2r << 3 | b2c;
												longest_loss[longest_losses][2] = wr << 3 | wc;
											 }
											longest_losses++;
										 }
									 }
									else
									 {
										sum[8]--;
										white_sum[white_slice][8]--;
									 }
									sum[8]++;
									white_sum[white_slice][8]++;
									if ( (datum >= 'B') && (datum <= 'Z')
									  && (otm_datum >= 'B') && (otm_datum <= 'Z') )
									 {
										if (loss_wins < LOSS_WIN_LIMIT)
										 {
											loss_win[loss_wins][0] = b1r << 3 | b1c;
											loss_win[loss_wins][1] = b2r << 3 | b2c;
											loss_win[loss_wins][2] = wr << 3 | wc;
										 }
										loss_wins++;
									 }
									if ( (datum >= 'B') && (datum <= 'Z')
									  && (otm_datum == '~') )
									 {
										if (loss_draws < LOSS_DRAW_LIMIT)
										 {
											loss_draw[loss_draws][0] = b1r << 3 | b1c;
											loss_draw[loss_draws][1] = b2r << 3 | b2c;
											loss_draw[loss_draws][2] = wr << 3 | wc;
										 }
										loss_draws++;
									 }
									if ( (datum == '~')
									  && (otm_datum >= 'B') && (otm_datum <= 'Z') )
									 {
										if (draw_wins < DRAW_WIN_LIMIT)
										 {
											draw_win[draw_wins][0] = b1r << 3 | b1c;
											draw_win[draw_wins][1] = b2r << 3 | b2c;
											draw_win[draw_wins][2] = wr << 3 | wc;
										 }
										draw_wins++;
									 }
								 }
							 }
						 }
						if (sum[8] != 0)
						 {
							if (blank_line_needed)
							 {
								fprintf(log_file, "\n");
								blank_line_needed = FALSE;
							 }
							fprintf(log_file, "%c%c%c/%c%c%c",
								LETTER[m1], b1c + 'a', b1r + '1',
								LETTER[m2], b2c + 'a', b2r + '1');
							for (count = 0; count < 9; count++)
							 {
								if (sum[count] > 0)
									fprintf(log_file, " %5u", sum[count]);
								else
									fprintf(log_file, "      ");
								pass_sum[pass][count] += sum[count];
							 }
							if (max_win > 0)
								fprintf(log_file, "    %2u %c%c%c", max_win,
									LETTER[w_man],
									max_win_col + 'a', max_win_row + '1');
							else
								fprintf(log_file, "          ");
							if (max_loss > 0)
								fprintf(log_file, "  %2u %c%c%c", max_loss,
									LETTER[w_man],
									max_loss_col + 'a', max_loss_row + '1');
							else
								fprintf(log_file, "        ");
							fprintf(log_file, "\n");
						 }
					 }
				 }
			 }
		 }
		fprintf(log_file, "\n");
		fprintf(log_file, "Total  ");
		for (count = 0; count < 9; count++)
			fprintf(log_file, " %5u", pass_sum[pass][count]);
		fprintf(log_file, "\n");
	 }

	if (passes > 1)
	 {
		fprintf(log_file, "\n");
		fprintf(log_file, "Both   ");
		for (count = 0; count < 9; count++)
			fprintf(log_file, " %5u", pass_sum[0][count] + pass_sum[1][count]);
		fprintf(log_file, "\n");
	 }

	fprintf(log_file, "\n");
	fprintf(log_file, "Analysis by position of White man\n");
	fprintf(log_file, "\n");
	fprintf(log_file, "            --- Wins ---       Draws      -- Losses --  Total\n");
	fprintf(log_file, "            1     2    3+     1    2+     1     2    3+      \n");

	for (count = 0; count < 9; count++)
		grand_sum[count] = 0;
	for (wr = 0; wr < 4; wr++)
	 {
		fprintf(log_file, "\n");
		for (wc = wr; wc < 4; wc++)
		 {
			white_slice = slice_map[wr][wc];
			fprintf(log_file, "%c%c%c    ",
				LETTER[w_man], wc + 'a', wr + '1');
			for (count = 0; count < 9; count++)
			 {
				if (white_sum[white_slice][count] > 0)
					fprintf(log_file, " %5u", white_sum[white_slice][count]);
				else
					fprintf(log_file, "      ");
				grand_sum[count] += white_sum[white_slice][count];
			 }
			fprintf(log_file, "\n");
		 }
	 }
	fprintf(log_file, "\n");
	fprintf(log_file, "Total  ");
	for (count = 0; count < 9; count++)
		fprintf(log_file, " %5u", grand_sum[count]);
	fprintf(log_file, "\n");

	fprintf(log_file, "\n");
	fprintf(log_file, "Analysis of wins and losses by length\n");
	title_needed = TRUE;
	for (count = 0; count < 54; count++)
	 {
		if (counts[0][count] + counts[1][count] > 0)
		 {
			if (title_needed)
			 {
				fprintf(log_file, "\n");
				if (count < 26)
					fprintf(log_file, "   Wins:   ", count + 1);
				else if (count == 26)
					fprintf(log_file, "   Draws:  ");
				else if (count < 53)
					fprintf(log_file, "   Losses: ");
				else
					fprintf(log_file, "   Total   ");
				title_needed = FALSE;
			 }
			else
				fprintf(log_file, "           ");
			if (count < 26)
				fprintf(log_file, "%2u", count + 1);
			else if (count == 26)
				fprintf(log_file, "  ");
			else if (count < 53)
				fprintf(log_file, "%2u", 53 - count);
			else
				fprintf(log_file, "  ");
			for (pass = 0; pass < passes; pass++)
			 {
				fprintf(log_file, "   %5u", counts[pass][count]);
				counts[pass][53] += counts[pass][count];
			 }
			fprintf(log_file, "\n");
		 }
		if ( (count == 25) || (count == 26) || (count == 52) )
			title_needed = TRUE;
	 }

	list_positions(draws, DRAW_LIMIT, draw, m1, m2,
		"non-trivial draws");
	list_positions(longest_wins, LONGEST_WIN_LIMIT, longest_win, m1, m2,
		"longest wins");
	list_positions(longest_losses, LONGEST_LOSS_LIMIT, longest_loss, m1, m2,
		"longest losses");
	list_positions(loss_wins, LOSS_WIN_LIMIT, loss_win, m1, m2,
		"non-trivial losses (wins with opponent to move)");
	list_positions(loss_draws, LOSS_DRAW_LIMIT, loss_draw, m1, m2,
		"non-trivial losses (draws with opponent to move)");
	list_positions(draw_wins, DRAW_WIN_LIMIT, draw_win, m1, m2,
		"non-trivial draws (wins with opponent to move)");

	if (analyses < ANALYSIS_LIMIT)
	 {
		men[analyses][0] = w_man;
		men[analyses][1] = m1;
		men[analyses][2] = m2;
		men[analyses][3] = wtm;
		for (count = 0; count < 8; count++)
			analysis_data[analyses][count] = grand_sum[count];
		analysis_data[analyses][8] = longest_win_so_far;
		analysis_data[analyses][9] = longest_loss_so_far;
		analysis_data[analyses][10] = loss_wins;
		analysis_data[analyses][11] = loss_draws;
		analysis_data[analyses][12] = draw_wins;
	 }
	analyses++;

	printf("\n");
	printf("\n");
	printf("   Analysis of %c v %c+%c, %s to move\n",
		LETTER[w_man], LETTER[m1], LETTER[m2],
		(wtm) ? "White" : "Black");
	printf("\n");
	printf("      Wins:    1                                      %5u\n", grand_sum[0]);
	printf("               2                                      %5u\n", grand_sum[1]);
	printf("               3+                                     %5u\n", grand_sum[2]);
	printf("\n");
	printf("      Draws:   1                                      %5u\n", grand_sum[3]);
	printf("               2+                                     %5u\n", grand_sum[4]);
	printf("\n");
	printf("      Losses:  1                                      %5u\n", grand_sum[5]);
	printf("               2                                      %5u\n", grand_sum[6]);
	printf("               3+                                     %5u\n", grand_sum[7]);
	printf("\n");
	printf("      Total                                           %5u\n", grand_sum[8]);
	printf("\n");
	printf("      Non-trivial losses, wins with opponent to move  %5u\n", loss_wins);
	printf("      Non-trivial losses, draws with opponent to move %5u\n", loss_draws);
	printf("      Non-trivial draws, wins with opponent to move   %5u\n", draw_wins);
	printf("\n");
	printf("   Full analysis written to log file\n");
 }

void analyse_white_position()
 {
	int slice, position, datum, value, non_trivial_slice, non_trivial_slices,
		b1r, b1c, b2r, b2c, m1, m2;
	char data[8][8];

	printf("\n");
	printf("   Analysing...");

	if (order_of_b_men == NORMAL)
	 {
		m1 = b1_man;
		m2 = b2_man;
	 }
	else
	 {
		m1 = b2_man;
		m2 = b1_man;
	 }
	fprintf(log_file, "\n");
	fprintf(log_file, "Analysis of %c v %c+%c for various positions of b%c, %s to move\n",
		LETTER[w_man], LETTER[m1], LETTER[m2], LETTER[m2],
		(wtm) ? "White" : "Black");

	non_trivial_slices = 0;
	for (b1r = 7; b1r >= 0; b1r--)
	 {
		for (b1c = 0; b1c < 8; b1c++)
		 {
			non_trivial_slice = FALSE;
			for (b2r = 7; b2r >= 0; b2r--)
			 {
				for (b2c = 0; b2c < 8; b2c++)
				 {
					if (order_of_b_men == NORMAL)
						encode_address(&slice, &position,
							b1r, b1c, b2r, b2c, w_row, w_col);
					else
						encode_address(&slice, &position,
							b2r, b2c, b1r, b1c, w_row, w_col);
					datum = (wtm) ? wtm_data[slice][position]
									  : btm_data[slice][position];
					data[b2r][b2c] = datum;
					if ( ( (datum >= 'b') && (datum <= 'z') )
					  || (datum == '~')
					  || ( (datum >= 'B') && (datum <= 'Z') ) )
						non_trivial_slice = TRUE;
				 }
			 }
			if (non_trivial_slice)
			 {
				fprintf(log_file, "\n");
				for (b2r = 7; b2r >= 0; b2r--)
				 {
					fprintf(log_file, "   ");
					for (b2c = 0; b2c < 8; b2c++)
					 {
						datum = data[b2r][b2c];
						if ( (datum >= 'A') && (datum <= 'Z') )
							value = - (datum - 'A' + 1);
						else if ( (datum == '~') || (datum == '|') )
							value = 0;
						else if (datum == '\\')
							value = 1;
						else if ( (datum >= 'a') && (datum <= 'z') )
							value = (datum - 'a' + 1);
						else
							value = - 99;
						if (value == - 99)
						 {
							if ( (b2r == w_row) && (b2c == w_col) )
								fprintf(log_file, "  w%c", LETTER[w_man]);
							else if ( (b2r == b1r) && (b2c == b1c) )
								fprintf(log_file, "  b%c", LETTER[m1]);
							else
								fprintf(log_file, "   *");
						 }
						else if (value == - 1)
							fprintf(log_file, "   .");
						else if (value == 0)
							fprintf(log_file, "   =");
						else
							fprintf(log_file, " %3d", value);
					 }
					fprintf(log_file, "\n");
				 }
				non_trivial_slices++;
			 }
		 }
	 }

	fprintf(log_file, "\n");
	if (non_trivial_slices == 0)
		fprintf(log_file, "No non-trivial positions\n");
	else
		fprintf(log_file, "Other positions are trivial\n");

	printf("\n");
	printf("\n");
	printf("   Analysis written to log file\n");
 }

int board_square(int row, int column)
 {
	return(BOARD_START + row * ROW_INC + column);
 }

void calculate_moves(int man, int row, int col, int man_number)
 {
	int square, next_square, increment, *increment_ptr, ok_to_repeat;

	square = board_square(row, col);
	increment_ptr = move_power[man];
	while ( (increment = *increment_ptr++) != 0)
	 {
		next_square = square;
		ok_to_repeat = (move_type[man] == RIDER);
		do
		 {
			next_square += increment;
			if (board[next_square] == EMPTY)
				target_board[next_square] |= man_number;
			else
				ok_to_repeat = FALSE;
		 } while (ok_to_repeat);
	 }
 }

void clear_board()
 {
	int count;

	for (count = 0; count < BOARD_ARRAY_SIZE; count++)
		board[count] = NOT_ON_BOARD;
 }

void clear_target_board()
 {
	int count;

	for (count = 0; count < BOARD_ARRAY_SIZE; count++)
		target_board[count] = 0;
 }

int colour_test(int b1r, int b1c, int b2r, int b2c)
 {
	if ( ( (LETTER[b1_man] == 'B') && (LETTER[b2_man] == 'B') )
	  || ( (LETTER[b1_man] == 'B') && (LETTER[b2_man] == 'N') )
	  || ( (LETTER[b1_man] == 'N') && (LETTER[b2_man] == 'B') ) )
		return(1 - ( (b1r + b1c + b2r + b2c) & 1));
	else
		return(0);
 }

void encode_address(int *slice, int *address,
	int b1_row, int b1_col, int b2_row, int b2_col, int w_row, int w_col)
 {
	int swap;

	if (b1_row < 4)
	 {
		b1_row = 7 - b1_row;
		b2_row = 7 - b2_row;
		w_row = 7 - w_row;
	 }
	if (b1_col > 3)
	 {
		b1_col = 7 - b1_col;
		b2_col = 7 - b2_col;
		w_col = 7 - w_col;
	 }
	swap = FALSE;
	if (b1_row + b1_col < 7)
		swap = TRUE;
	else if (b1_row + b1_col == 7)
	 {
		if (b2_row + b2_col < 7)
			swap = TRUE;
		else if (b2_row + b2_col == 7)
		 {
			if (w_row + w_col < 7)
				swap = TRUE;
		 }
	 }
	if (swap)
	 {
		swap = b1_row;
		b1_row = 7 - b1_col;
		b1_col = 7 - swap;
		swap = b2_row;
		b2_row = 7 - b2_col;
		b2_col = 7 - swap;
		swap = w_row;
		w_row = 7 - w_col;
		w_col = 7 - swap;
	 }

	*slice = slice_map[b1_row][b1_col];
	*address = ( ( ( ( (b2_row << 3) | b2_col) << 3) | w_row) << 3) | w_col;
 }

int find_char(char *string, int c)
 {
	int count;

	count = 0;
	while (string[count] != 0)
	 {
		if (string[count] == c)
			return(count);
		count++;
	 }
	return(-1);
 }

int good_address(int b1r, int b1c, int b2r, int b2c, int wr, int wc)
 {
	int new_wr, new_wc, new_b1r, new_b1c, new_b2r, new_b2c, swap;

	if ( (b1r == wr) && (b1c == wc) )
		return(FALSE);
	if ( (wr == b2r) && (wc == b2c) )
		return(FALSE);
	if ( (b1r == b2r) && (b1c == b2c) )
		return(FALSE);

	if ( (b1r + b1c == 7) && (b2r + b2c < 7) )
		return(FALSE);

	if ( (b1r + b1c == 7) && (b2r + b2c == 7) && (wr + wc < 7) )
		return(FALSE);

	if (b1_man != b2_man)
		return(TRUE);
	else
	 {
		new_wr = wr;
		new_wc = wc;
		new_b1r = b2r;
		new_b1c = b2c;
		new_b2r = b1r;
		new_b2c = b1c;
		if (new_b1r < 4)
		 {
			new_wr = 7 - new_wr;
			new_b1r = 7 - new_b1r;
			new_b2r = 7 - new_b2r;
		 }
		if (new_b1c >= 4)
		 {
			new_wc = 7 - new_wc;
			new_b1c = 7 - new_b1c;
			new_b2c = 7 - new_b2c;
		 }
		swap = FALSE;
		if (new_b1r + new_b1c < 7)
			swap = TRUE;
		else if (new_b1r + new_b1c == 7)
		 {
			if (new_b2r + new_b2c < 7)
				swap = TRUE;
			else if (new_b2r + new_b2c == 7)
			 {
				if (new_wr + new_wc < 7)
					swap = TRUE;
			 }
		 }
		if (swap)
		 {
			swap = new_wr;
			new_wr = 7 - new_wc;
			new_wc = 7 - swap;
			swap = new_b1r;
			new_b1r = 7 - new_b1c;
			new_b1c = 7 - swap;
			swap = new_b2r;
			new_b2r = 7 - new_b2c;
			new_b2c = 7 - swap;
		 }
		if (b1r > new_b1r)
			return(TRUE);
		else if (b1r < new_b1r)
			return(FALSE);
		if (b1c < new_b1c)
			return(TRUE);
		else if (b1c > new_b1c)
			return(FALSE);
		if (b2r > new_b2r)
			return(TRUE);
		else if (b2r < new_b2r)
			return(FALSE);
		if (b2c < new_b2c)
			return(TRUE);
		else if (b2c > new_b2c)
			return(FALSE);
		if (wr > new_wr)
			return(TRUE);
		else if (wr < new_wr)
			return(FALSE);
		if (wc < new_wc)
			return(TRUE);
		else if (wc > new_wc)
			return(FALSE);
		return(TRUE);
	 }
 }

void list_and_print_moves()
 {
	int man, old_row, old_col, row, col,
		square, pass, passes, count, slice, position,
		man_to_move, datum, new_result, shortest_win, draw_found, longest_loss,
		lines, columns, line_count, column_count;

	moves = 0;
	shortest_win = 100;
	draw_found = FALSE;
	longest_loss = 0;
	passes = (wtm) ? 1 : 2;

	for (pass = 1; pass <= passes;  pass++)
	 {
		if (wtm)
		 {
			man = w_man;
			old_row = w_row;
			old_col = w_col;
		 }
		else if (pass == 1)
		 {
			man = b1_man;
			old_row = b1_row;
			old_col = b1_col;
		 }
		else
		 {
			man = b2_man;
			old_row = b2_row;
			old_col = b2_col;
		 }
		for (row = 7; row >= 0; row--)
		 {
			for (col = 0; col < 8; col++)
			 {
				square = board_square(row, col);
				if ( (target_board[square] & pass) != 0)
				 {
					if (wtm)
					 {
						man_to_move = W_MAN;
						encode_address(&slice, &position,
							b1_row, b1_col, b2_row, b2_col, row, col);
						datum = btm_data[slice][position];
					 }
					else
					 {
						if (pass == 1)
						 {
							man_to_move = B1_MAN;
							encode_address(&slice, &position,
								row, col, b2_row, b2_col, w_row, w_col);
						 }
						else
						 {
							man_to_move = B2_MAN;
							encode_address(&slice, &position,
								b1_row, b1_col, row, col, w_row, w_col);
						 }
						datum = wtm_data[slice][position];
					 }
					if ( (datum >= 'a') && (datum <= 'z') )
						new_result = - (datum - 'a' + 1);
					else if ( (datum >= 'A') && (datum <= 'Z') )
						new_result = datum - 'A' + 1;
					else if ( (datum == '|') || (datum == '~') )
					 {
						new_result = 0;
						draw_found = TRUE;
					 }
					else if (datum == '\\')
						new_result = - 1;
					else
						new_result = - 99;
					if ( (new_result > 0) && (new_result < shortest_win) )
						shortest_win = new_result;
					if ( (new_result < 0) && (new_result != - 99)
												 && (new_result < longest_loss) )
						longest_loss = new_result;
					move[moves][0] = man_to_move;
					move[moves][1] = man;
					move[moves][2] = old_row;
					move[moves][3] = old_col;
					move[moves][4] = row;
					move[moves][5] = col;
					move[moves][6] = new_result;
					moves++;
				 }
			 }
		 }
	 }

	if (shortest_win < 100)
		best_result = shortest_win;
	else if (draw_found)
		best_result = 0;
	else
		best_result = longest_loss;
	best_move = 0;
	unique_best_move = TRUE;
	for (count = moves - 1; count >= 0; count--)
	 {
		if (move[count][6] == best_result)
		 {
			move[count][7] = '*';
			if (best_move != 0)
				unique_best_move = FALSE;
			best_move = count + 1;
		 }
		else
			move[count][7] = ' ';
	 }

	if (moves == 0)
	 {
		printf("\n");
		printf("No moves available\n");
	 }
	else
	 {
		printf("\n");
		fprintf(log_file, "\n");
		lines = (moves > 15) ? 15 : moves;
		columns = ( (moves - 1) / 15) + 1;
		for (line_count = 0; line_count < lines; line_count++)
		 {
			count = line_count;
			for (column_count = 0; column_count < columns; column_count++)
			 {
				if (count < moves)
				 {
					printf("   %2u %c%c%c-%c%c %3d %c ",
						count + 1, LETTER[move[count][1]],
						move[count][3] + 'a',
						move[count][2] + '1',
						move[count][5] + 'a',
						move[count][4] + '1',
						move[count][6],
						move[count][7]);
					fprintf(log_file, "   %2u %c%c%c-%c%c %3d %c ",
						count + 1, LETTER[move[count][1]],
						move[count][3] + 'a',
						move[count][2] + '1',
						move[count][5] + 'a',
						move[count][4] + '1',
						move[count][6],
						move[count][7]);
				 }
				count += lines;
			 }
			printf("\n");
			fprintf(log_file, "\n");
		 }
	 }
 }

void list_moves_played()
 {
	int move, move_count, black_starts, half_move, last_half_move,
		lines, line_count, columns, column_count, slice, position, datum;

	if (moves_played == 0)
	 {
		printf("\n");
		printf("   No moves played\n");
	 }
	else
	 {
		black_starts = !move_played[0][0];
		printf("\n");
		printf("   Opening position %s, %s to move\n",
			opening_position, (black_starts) ? "Black" : "White");
		fprintf(log_file, "\n");
		fprintf(log_file, "Opening position %s, %s to move\n",
			opening_position, (black_starts) ? "Black" : "White");
		last_half_move = moves_played;
		if (black_starts)
			last_half_move++;
		lines = (last_half_move + 1) / 2;
		columns = 2;
		lines = (lines + 1) / 2;

		printf("\n");
		fprintf(log_file, "\n");
		for (line_count = 0; line_count < lines; line_count++)
		 {
			half_move = line_count * 2;
			if (!black_starts) half_move++;
			for (column_count = 0; column_count < columns; column_count++)
			 {
				if (half_move <= moves_played)
				 {
					printf("       %2u  ", line_count + 1 + column_count * lines);
					fprintf(log_file, "       %2u  ", line_count + 1 + column_count * lines);
					if (half_move == 0)
					 {
						printf("......      ");
						fprintf(log_file, "......      ");
					 }
					else
					 {
						printf("%c%c%c-%c%c %c    ",
							LETTER[move_played[half_move][7]],
							move_played[half_move][9] + 'a',
							move_played[half_move][8] + '1',
							move_played[half_move][11] + 'a',
							move_played[half_move][10] + '1',
							move_played[half_move][12]);
						fprintf(log_file, "%c%c%c-%c%c %c    ",
							LETTER[move_played[half_move][7]],
							move_played[half_move][9] + 'a',
							move_played[half_move][8] + '1',
							move_played[half_move][11] + 'a',
							move_played[half_move][10] + '1',
							move_played[half_move][12]);
					 }
					half_move++;
					if (half_move <= moves_played)
					 {
						printf("%c%c%c-%c%c %c    ",
							LETTER[move_played[half_move][7]],
							move_played[half_move][9] + 'a',
							move_played[half_move][8] + '1',
							move_played[half_move][11] + 'a',
							move_played[half_move][10] + '1',
							move_played[half_move][12]);
						fprintf(log_file, "%c%c%c-%c%c %c    ",
							LETTER[move_played[half_move][7]],
							move_played[half_move][9] + 'a',
							move_played[half_move][8] + '1',
							move_played[half_move][11] + 'a',
							move_played[half_move][10] + '1',
							move_played[half_move][12]);
					 }
					half_move += (2 * lines - 1);
				 }
			 }
			printf("\n");
			fprintf(log_file, "\n");
		 }
		encode_address(&slice, &position,
			b1_row, b1_col, b2_row, b2_col, w_row, w_col);
		if (wtm)
			datum = wtm_data[slice][position];
		else
			datum = btm_data[slice][position];
		printf("\n");
		printf("   ");
		fprintf(log_file, "\n");
		fprintf(log_file, "   ");
		print_and_echo_result(datum);

		printf("\n");
		printf("   ! indicates a unique optimal move, ? a non-optimal move\n");
		fprintf(log_file, "\n");
		fprintf(log_file, "   ! indicates a unique optimal move, ? a non-optimal move\n");
	 }
 }

void list_positions(int items, int limit, unsigned char item[][3],
	int m1, int m2, char *title)
 {
	int count, last_b1, last_b2, datum;

	fprintf(log_file, "\n");
	if (items == 0)
		fprintf(log_file, "No %s\n", title);
	else if (items > limit)
		fprintf(log_file, "More than %u %s\n", limit, title);
	else
	 {
		fprintf(log_file, "List of %s", title);
		last_b1 = item[0][0] + 1;
		for (count = 0; count < items; count++)
		 {
			if (last_b1 != item[count][0])
			 {
				fprintf(log_file, "\n");
				last_b1 = item[count][0];
				last_b2 = item[count][1] + 1;
			 }
			if (last_b2 != item[count][1])
			 {
				fprintf(log_file, "\n");
				fprintf(log_file, "   %c%c%c/%c%c%c v %c",
					LETTER[m1],
					(item[count][0] & 7) + 'a', (item[count][0] >> 3) + '1',
					LETTER[m2],
					(item[count][1] & 7) + 'a', (item[count][1] >> 3) + '1',
					LETTER[w_man]);
				last_b2 = item[count][1];
			 }
			else
				fprintf(log_file, ",");
			fprintf(log_file, "%c%c",
				(item[count][2] & 7) + 'a', (item[count][2] >> 3) + '1');
		 }
		fprintf(log_file, "\n");
	 }
 }

int print_and_echo_result(int datum)
 {
	int length;
	char *player_to_move = (wtm) ? "White" : "Black";

	length = 0;
	if ( (datum >= 'a') && (datum <= 'z') )
	 {
		length = datum - 'a' + 1;
		printf("%s wins in %u move", player_to_move, length);
		fprintf(log_file, "%s wins in %u move", player_to_move, length);
	 }
	else if ( (datum >= 'B') && (datum <= 'Z') )
	 {
		length = datum - 'A' + 1;
		printf("%s loses in %u move", player_to_move, length);
		fprintf(log_file, "%s loses in %u move", player_to_move, length);
	 }
	else if (datum == 'A')
	 {
		printf("%s loses by immediate capture", player_to_move);
		fprintf(log_file, "%s loses by immediate capture", player_to_move);
	 }
	else if (datum == '|')
	 {
		printf("%s draws by immediate capture", player_to_move);
		fprintf(log_file, "%s draws by immediate capture", player_to_move);
	 }
	else if (datum == '\\')
	 {
		printf("%s wins by immediate capture", player_to_move);
		fprintf(log_file, "%s wins by immediate capture", player_to_move);
	 }
	else if (datum == '~')
	 {
		length = 1;
		printf("Draw");
		fprintf(log_file, "Draw");
	 }
	else
	 {
		printf("result not recorded");
		fprintf(log_file, "result not recorded");
	 }
	if (length > 1)
	 {
		printf("s");
		fprintf(log_file, "s");
	 }
	printf("\n");
	fprintf(log_file, "\n");

	return(length);
 }

void print_help_screen()
 {
	printf("\n");
	printf("   Instructions for use of LOSING3 (screen 1 of 2)\n");
	printf("\n");
	printf("   To set a position, give White man, Black man 1, Black man 2, side to move\n");
	printf("\n");
	printf("   For example, to set wBd2 against bKd5 and bRd7, Black to move, type\n");
	printf("\n");
	printf("       Bd2 Kd5 Rd7 B\n");
	printf("\n");
	printf("   Pieces are K Q R B N, letters may be in either case, spaces are ignored\n");
	printf("       but input is limited to 20 characters\n");
	printf("\n");
	printf("   If side to move is omitted, White is assumed\n");
	printf("\n");
	printf("   When a list of available moves is displayed, type n to choose move n\n");
	printf("       (ENTER alone selects the best move)\n");
	printf("\n");
	printf("   If moves have already been played, type -n to retract the last n moves\n");
	printf("       (- alone retracts a single move)\n");
	printf("\n");
	printf("   Type L to list the moves made so far, P to repeat the current position\n");
	printf("       and the moves available from it, S to print the table of statistics\n");
	printf("       (see next screen), X to exit from the program\n");
	printf("\n");

	printf("Press ENTER to see screen 2: ");
	console_input_buffer[0] = 1;
	cgets(console_input_buffer);
	printf("\n");

	printf("\n");
	printf("   Instructions for use of LOSING3 (screen 2 of 2)\n");
	printf("\n");
	printf("   Use ? to obtain simultaneous data for several positions\n");
	printf("\n");
	printf("   It is possible to apply ? to the White man, the Black men, or all three\n");
	printf("\n");
	printf("   Examples (using R v N+N):\n");
	printf("\n");
	printf("      R ? Nd5 Ne5\n");
	printf("\n");
	printf("   to display all results with bNd5 and bNe5 (WTM and BTM appear side by side),\n");
	printf("\n");
	printf("      Rh1 N ? N B\n");
	printf("\n");
	printf("   to report all results with wRh1 and Black to move (omit B for WTM), and\n");
	printf("\n");
	printf("      R ? N ? N\n");
	printf("\n");
	printf("   to summarize all possible results with White to move (add B for BTM)\n");
	printf("\n");
	printf("   In the last case, the most important statistics are added to an internal\n");
	printf("      table, and a subsequent command S prints this table\n");
 }

void print_position_and_moves()
 {
	int slice, position, datum, length;

	char *player_to_move;

	encode_address(&slice, &position,
		b1_row, b1_col, b2_row, b2_col, w_row, w_col);
	if (wtm)
		datum = wtm_data[slice][position];
	else
		datum = btm_data[slice][position];
	player_to_move = (wtm) ? "White" : "Black";
	set_position(current_position, b1_row, b1_col, b2_row, b2_col, w_row, w_col);
	printf("\n");
	printf("%s, %s to move:  ", current_position, player_to_move);
	fprintf(log_file, "\n");
	fprintf(log_file, "%s, %s to move:  ", current_position, player_to_move);
	length = print_and_echo_result(datum);

	set_board();
	clear_target_board();
	if (length >= 1)
	 {
		if (wtm)
			calculate_moves(w_man, w_row, w_col, 1);
		else
		 {
			calculate_moves(b1_man, b1_row, b1_col, 1);
			calculate_moves(b2_man, b2_row, b2_col, 2);
		 }
		list_and_print_moves();
	 }
	else
		moves = 0;
 }

void read_and_obey_console_command()
 {
	int c, piece, input_state, count, man_count, action, number, digits_read,
		error_found, error_pointer;
	int man[3], row[3], col[3], side_to_move;
	int slice, position, datum, pass, value, length;

	char *error_message, *player_to_move;

	error_found = FALSE;

	printf("\n");
	if (moves > 0)
		printf("Give move number or new position (H for help), then ENTER: ");
	else
		printf("Give position (H for help, X to exit), then press ENTER: ");
	console_input_buffer[0] = CONSOLE_INPUT_LIMIT + 1;
	cgets(console_input_buffer);
	printf("\n");

	input_count = console_input_buffer[1];

/* in Turbo C, the test below is unnecessary since "cgets" never returns an
	overlength string, but it has been kept as a possible aid to portability */

	if (input_count <= CONSOLE_INPUT_LIMIT)
	 {
		console_input_buffer[input_count + 2] = ARTIFICIAL_INPUT_TERMINATOR;
		console_input_buffer[input_count + 3] = 0;
	 }
	else
	 {
		error_pointer = CONSOLE_INPUT_LIMIT;
		error_message = "Too many characters (limit 20)";
		error_found = TRUE;
	 }

	input_count = 2;
	w_queried = FALSE;
	b_queried = FALSE;
	side_to_move = 'W';
	input_state = READING_FIRST_CHARACTER;
	while ( (!error_found)
		  && ( (c = console_input_buffer[input_count++]) != 0) )
	 {
		if (c > ' ')
		 {
			switch (input_state)
			 {
				case READING_FIRST_CHARACTER:

					c = toupper(c);
					piece = find_char(LETTER, c);

					if (piece >= 0)
					 {
						man_count = 0;
						man[0] = piece;
						input_state = READING_COL;
					 }
					else if ( (c >= '0') && (c <= '9') )
					 {
						if (moves > 0)
						 {
							action = CHOOSE_MOVE;
							number = c - '0';
							digits_read = 1;
							input_state = READING_INTEGER;
						 }
						else
						 {
							error_pointer = input_count - 3;
							error_message = "No moves listed";
							error_found = TRUE;
						 }
					 }
					else if (c == ARTIFICIAL_INPUT_TERMINATOR)
					 {
						if (moves > 0)
						 {
							action = CHOOSE_MOVE;
							number = 0;
							input_state = SKIPPING_TO_END;
						 }
						else
							action = DO_NOTHING;
					 }
					else if (c == '-')
					 {
						if (moves_played > 0)
						 {
							action = RETRACT_MOVES;
							number = 0;
							digits_read = 0;
							input_state = READING_INTEGER;
						 }
						else
						 {
							error_pointer = input_count - 3;
							error_message = "No moves played";
							error_found = TRUE;
						 }
					 }
					else if (c == 'H')
					 {
						action = GIVE_HELP;
						input_state = SKIPPING_TO_END;
					 }
					else if (c == 'L')
					 {
						action = LIST_MOVES_PLAYED;
						input_state = SKIPPING_TO_END;
					 }
					else if (c == 'P')
					 {
						action = REPEAT_POSITION;
						input_state = SKIPPING_TO_END;
					 }
					else if (c == 'S')
					 {
						action = REPORT_STATISTICS;
						input_state = SKIPPING_TO_END;
					 }
					else if (c == 'X')
					 {
						action = EXIT;
						input_state = SKIPPING_TO_END;
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Input not understood";
						error_found = TRUE;
					 }

					break;

				case READING_MAN:

					c = toupper(c);
					piece = find_char(LETTER, c);

					if (piece >= 0)
					 {
						man[man_count] = piece;
						if (b_queried)
						 {
							action = OBEY_QUERIES;
							input_state = READING_SIDE_TO_MOVE;
						 }
						else
							input_state = READING_COL;
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Man not one of K N B R Q";
						error_found = TRUE;
					 }

					break;

				case READING_COL:

					c = tolower(c);
					if ( (c >= 'a') && (c <= 'h') )
					 {
						col[man_count] = c - 'a';
						input_state = READING_ROW;
					 }
					else if (c == '?')
					 {
						col[man_count] = - 1;
						row[man_count] = - 1;
						if (man_count == 0)
						 {
							w_queried = TRUE;
							input_state = READING_MAN;
						 }
						else if (man_count == 1)
						 {
							b_queried = TRUE;
							input_state = READING_MAN;
						 }
						else
						 {
							error_pointer = input_count - 3;
							error_message = "Query not allowed here";
							error_found = TRUE;
						 }
						man_count++;
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Column not in range a-h";
						error_found = TRUE;
					 }

					break;

				case READING_ROW:

					if ( (c >= '1') && (c <= '8') )
					 {
						row[man_count] = c - '1';
						if (man_count > 0)
						 {
							for (count = 0; count < man_count; count++)
							 {
								if ( (row[count] == row[man_count])
								  && (col[count] == col[man_count]) )
								 {
									error_pointer = input_count - 3;
									error_message = "Men on the same square";
									error_found = TRUE;
								 }
							 }
						 }
						man_count++;
						if (man_count < 3)
							input_state = READING_MAN;
						else
						 {
							if (w_queried)
							 {
								action = OBEY_QUERIES;
								input_state = SKIPPING_TO_END;
							 }
							else
							 {
								action = SET_POSITION;
								input_state = READING_SIDE_TO_MOVE;
							 }
						 }
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Row not in range 1-8";
						error_found = TRUE;
					 }

					break;

				case READING_SIDE_TO_MOVE:

					c = toupper(c);
					if (c == ARTIFICIAL_INPUT_TERMINATOR)
						c = 'W';

					if ( (c == 'W') || (c == 'B') )
					 {
						side_to_move = c;
						input_state = SKIPPING_TO_END;
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Side to move must be W or B";
						error_found = TRUE;
					 }

					break;

				case READING_INTEGER:

					if (c == ARTIFICIAL_INPUT_TERMINATOR)
						input_state = SKIPPING_TO_END;
					else if ( (c >= '0') && (c <= '9') )
					 {
						if (digits_read >= 2)
						 {
							error_pointer = input_count - 3;
							error_message = "Too many digits";
							error_found = TRUE;
						 }
						else
						 {
							number = number * 10 + (c - '0');
							digits_read++;
						 }
					 }
					else
					 {
						error_pointer = input_count - 3;
						error_message = "Illegal character in number";
						error_found = TRUE;
					 }

					break;

				case SKIPPING_TO_END:

					break;

				default:

					break;
			 }
		 }
	 }

	if (error_found)
		action = REPORT_ERROR;

	if ( (action == SET_POSITION) || (action == OBEY_QUERIES) )
	 {
		w_man = man[0];
		w_row = row[0];
		w_col = col[0];
		if (man[1] >= man[2])
		 {
			b1_man = man[1];
			b1_row = row[1];
			b1_col = col[1];
			b2_man = man[2];
			b2_row = row[2];
			b2_col = col[2];
			order_of_b_men = NORMAL;
		 }
		else
		 {
			b1_man = man[2];
			b1_row = row[2];
			b1_col = col[2];
			b2_man = man[1];
			b2_row = row[1];
			b2_col = col[1];
			order_of_b_men = INVERTED;
		 }
		wtm = (side_to_move == 'W');
	 }

	switch(action)
	 {
		case SET_POSITION:

			set_position(opening_position,
				b1_row, b1_col, b2_row, b2_col, w_row, w_col);

			moves_played = 0;
			move_played[0][0] = wtm;
			move_played[0][1] = w_row;
			move_played[0][2] = w_col;
			move_played[0][3] = b1_row;
			move_played[0][4] = b1_col;
			move_played[0][5] = b2_row;
			move_played[0][6] = b2_col;

			if (read_database())
			 {
				position_set = TRUE;
				last_w_man = w_man;
				last_b1_man = b1_man;
				last_b2_man = b2_man;
				print_position_and_moves();
			 }
			else
			 {
				position_set = FALSE;
				last_w_man = - 1;
				moves = 0;
				if (brackets_read[w_man] >= 0)
				 {
					rewind(input_file[w_man]);
					brackets_read[w_man] = 0;
				 }
			 }

			break;

		case CHOOSE_MOVE:

			if (number == 0)
				number = best_move;
			if (moves_played >= MOVES_PLAYED_LIMIT)
			 {
				printf("\n");
				printf("%u moves already played\n", MOVES_PLAYED_LIMIT);
			 }
			else if (number > moves)
			 {
				printf("\n");
				printf("Not enough moves available\n");
			 }
			else
			 {
				printf("\n");
				printf("Move chosen: %u\n", number);
				fprintf(log_file, "\n");
				fprintf(log_file, "Move chosen: %u\n", number);
				number--;

				wtm = !wtm;
				switch(move[number][0])
				 {
					case W_MAN:

						w_row = move[number][4];
						w_col = move[number][5];
						break;

					case B1_MAN:

						b1_row = move[number][4];
						b1_col = move[number][5];
						break;

					case B2_MAN:

						b2_row = move[number][4];
						b2_col = move[number][5];
						break;

					default:

						break;
				 }

				moves_played++;
				move_played[moves_played][0] = wtm;
				move_played[moves_played][1] = w_row;
				move_played[moves_played][2] = w_col;
				move_played[moves_played][3] = b1_row;
				move_played[moves_played][4] = b1_col;
				move_played[moves_played][5] = b2_row;
				move_played[moves_played][6] = b2_col;
				move_played[moves_played][7] = move[number][1];
				move_played[moves_played][8] = move[number][2];
				move_played[moves_played][9] = move[number][3];
				move_played[moves_played][10] = move[number][4];
				move_played[moves_played][11] = move[number][5];
				if (move[number][7] != '*')
					move_played[moves_played][12] = '?';
				else if (unique_best_move)
					move_played[moves_played][12] = '!';
				else
					move_played[moves_played][12] = ' ';
			 }
			print_position_and_moves();

			break;

		case RETRACT_MOVES:

			if (moves_played == 0)
			 {
				printf("\n");
				printf("No moves to retract\n");
			 }
			else
			 {
				if (number == 0)
					number = 1;
				if (number > moves_played)
					number = moves_played;
				moves_played -= number;
				wtm = move_played[moves_played][0];
				w_row = move_played[moves_played][1];
				w_col = move_played[moves_played][2];
				b1_row = move_played[moves_played][3];
				b1_col = move_played[moves_played][4];
				b2_row = move_played[moves_played][5];
				b2_col = move_played[moves_played][6];

				print_position_and_moves();
			 }

			break;

		case OBEY_QUERIES:

			position_set = FALSE;
			moves = 0;
			if (read_database())
			 {
				last_w_man = w_man;
				last_b1_man = b1_man;
				last_b2_man = b2_man;
				if (w_queried && b_queried)
					analyse_complete_database();
				else if (w_queried)
					analyse_black_position();
				else
					analyse_white_position();
			 }
			else
			 {
				last_w_man = - 1;
				if (brackets_read[w_man] >= 0)
				 {
					rewind(input_file[w_man]);
					brackets_read[w_man] = 0;
				 }
			 }

			break;

		case GIVE_HELP:

			print_help_screen();

			break;

		case LIST_MOVES_PLAYED:

			list_moves_played();

			break;

		case REPEAT_POSITION:

			print_position_and_moves();

			break;

		case REPORT_STATISTICS:

			report_run_statistics();

			break;

		case DO_NOTHING:

			break;

		case REPORT_ERROR:

			if (moves > 0)
				printf("                                                           ");
			else
				printf("                                                         ");
			while (error_pointer > 0)
			 {
				printf(" ");
				error_pointer--;
			 }
			printf("*\n");
			printf("%s\n", error_message);

			break;

		case EXIT:

			quit = TRUE;

			break;

		default:

			printf("\n");
			printf("Sorry, not available yet\n");

			break;
	 }
 }

int read_database()
 {
	long check_sum, check_sum_read, values_expected, values_read;

	int input_state, c, pass, count, section_required,
		wr, wc, b1r, b1c, b2r, b2c, slice, position;

	char players[6], players_wanted[6];

	if ( (w_man == last_w_man) && (b1_man == last_b1_man)
										&& (b2_man == last_b2_man) )
		return(TRUE);
	else if (brackets_read[w_man] < 0)
	 {
		printf("\n");
		printf("Database file %s not available\n", input_filename[w_man]);
		return(FALSE);
	 }
	else
	 {
		printf("\n");
		printf("   Searching database file %s...", input_filename[w_man]);
		section_required = section_map[b1_man][b2_man];
		sprintf(players_wanted, "%cv%c+%c",
			LETTER[w_man], LETTER[b1_man], LETTER[b2_man]);
		if (section_required <= brackets_read[w_man])
		 {
			rewind(input_file[w_man]);
			brackets_read[w_man] = 0;
		 }
		input_state = FINDING_SECTION;

		while ( (c = fgetc(input_file[w_man])) != EOF)
		 {
			if (c > ' ')
			 {
				switch(input_state)
				 {
					case FINDING_SECTION:

						if (c == '[')
						 {
							brackets_read[w_man]++;
							if (section_required == brackets_read[w_man])
							 {
								count = 0;
								input_state = READING_PLAYERS;
							 }
							else
								printf(".");
						 }

						break;

					case READING_PLAYERS:

						if (c != ']')
						 {
							if (count < 5)
								players[count++] = c;
						 }
						else
						 {
							players[count] = 0;
							if (strcmp(players, players_wanted) != 0)
							 {
								printf("\n");
								printf("   Database corrupt, actual section is %s\n",
									players);
								return(FALSE);
							 }
							else
							 {
								values_expected = (b1_man == b2_man) ? 31444 : 62664;
								values_read = 0;
								check_sum = 0;
								printf("\n");
								printf("   Reading data, %lu values expected...",
									values_expected);
								for (slice = 0; slice < 10; slice++)
									for (position = 0; position < 4096; position++)
										wtm_data[slice][position] =
											btm_data[slice][position] = 0;
								pass = 0;
								b1r = b2r = wr = 7;
								b1c = b2c = wc = 0;
								count = 1;
								input_state = READING_RESULTS;
							 }
						 }

						break;

					case READING_RESULTS:

						if (c == '{')
						 {
							printf("\n");
							printf("   Values read:  %5lu\n", values_read);
							if (values_read != values_expected)
							 {
								printf("   Database corrupt\n");
								return(FALSE);
							 }
							check_sum_read = 0;
							input_state = READING_CHECK_SUM;
						 }
						else
						 {
							check_sum += c;
							if ( (c >= 33) && (c <= 51) )
								count = c - 30;
							else if ( ( (c >= 52) && (c <= 63) )
									 || ( (c >= 'a') && (c <= 'z') )
									 || ( (c >= 'A') && (c <= 'Z') )
									 || (c == '\\') || (c == '|') || (c == '~') )
							 {
								if ( (c >= 52) && (c <= 57) )
								 {
									count = c - 50;
									c = 'A';
								 }
								else if ( (c >= 58) && (c <= 63) )
								 {
									count = c - 56;
									c = 'a';
								 }
								while (count > 0)
								 {
									do
									 {
										wc++;
										if (wc > 7)
										 {
											wc = 0;
											wr--;
											if (wr < 0)
											 {
												wr = 7;
												b2c++;
												if (b2c > 7)
												 {
													b2c = 0;
													b2r--;
													if (b2r < 0)
													 {
														b2r = 7;
														b1c++;
														if (b1c > 3)
														 {
															b1r--;
															b1c = 7 - b1r;
															if (b1r < 4)
															 {
																b1r = 7;
																b1c = 0;
																pass++;
																if (pass > 1)
																 {
																	printf("\n");
																	printf("   Database corrupt, too many data values\n");
																	return(FALSE);
																 }
															 }
														 }
													 }
												 }
											 }
										 }
									 } while (! good_address(b1r, b1c, b2r, b2c,
											wr, wc));
									encode_address(&slice, &position,
										b1r, b1c, b2r, b2c, wr, wc);
									if (pass == 0)
										wtm_data[slice][position] = c;
									else
										btm_data[slice][position] = c;
									if (b1_man == b2_man)
									 {
										encode_address(&slice, &position,
											b2r, b2c, b1r, b1c, wr, wc);
										if (pass == 0)
											wtm_data[slice][position] = c;
										else
											btm_data[slice][position] = c;
									 }
									values_read++;
									count--;
								 }
								count = 1;
							 }
							else
							 {
								printf("\n");
								printf("   Database corrupt, unexpected character %c found\n", c);
								return(FALSE);
							 }
						 }

						break;

					case READING_CHECK_SUM:

						if ( (c >= '0') && (c <= '9') )
							check_sum_read = (check_sum_read * 10) + (c - '0');
						else if (c == '}')
						 {
							if (check_sum == check_sum_read)
								return(TRUE);
							else
							 {
								printf("   Database corrupt, sum check not satisfied\n");
								return(FALSE);
							 }
						 }
						else
						 {
							printf("   Database corrupt, character %c in check sum\n", c);
							return(FALSE);
						 }

						break;

					default:

						break;
				 }
			 }
		 }
		printf("\n");
		printf("   Database corrupt, section incomplete\n");
		return(FALSE);
	 }
 }

void report_run_statistics()
 {
	int count, limit, column_count, value;

	printf("\n");
	if (analyses == 0)
		printf("   No analyses performed\n");
	else
	 {
		fprintf(log_file, "\n");
		fprintf(log_file, "Summary of analyses\n");
		fprintf(log_file, "\n");
		fprintf(log_file, "- Men -  T      --- Wins ---       Draws      -- Losses --    Max   Max  Loss  Loss  Draw\n");
		fprintf(log_file, "W   -B-  M      1     2    3+     1    2+     1     2    3+   win  loss  /win /draw  /win\n");
		limit = (analyses > ANALYSIS_LIMIT) ? ANALYSIS_LIMIT : analyses;
		for (count = 0; count < limit; count++)
		 {
			if (count % 5 == 0)
				fprintf(log_file, "\n");
			fprintf(log_file, "%c v %c+%c  %c ", LETTER[men[count][0]],
				LETTER[men[count][1]], LETTER[men[count][2]],
				(men[count][3]) ? 'W' : 'B');
			for (column_count = 0; column_count < 13; column_count++)
			 {
				value = analysis_data[count][column_count];
				if (value > 0)
					fprintf(log_file, " %5u", value);
				else
					fprintf(log_file, "      ");
			 }
			fprintf(log_file, "\n");
		 }
		if (analyses > ANALYSIS_LIMIT)
		 {
			fprintf(log_file, "\n");
			fprintf(log_file, "More than %u analyses performed\n", ANALYSIS_LIMIT);
		 }
		printf("   Analyses performed: %u\n", analyses);
		printf("\n");
		printf("   Statistics written to log file\n");
	 }
 }

void set_board()
 {
	int row, column, square;

	for (row = 0; row < 8; row++)
		for (column= 0; column < 8; column++)
			board[board_square(row, column)] = EMPTY;

	board[board_square(w_row, w_col)] = w_man;
	board[board_square(b1_row, b1_col)] = b1_man;
	board[board_square(b2_row, b2_col)] = b2_man;
 }

void set_position(char *string,
	int b1r, int b1c, int b2r, int b2c, int wr, int wc)
 {
	string[0] = LETTER[w_man];
	string[1] = wc + 'a';
	string[2] = wr + '1';
	string[3] = ' ';
	string[4] = 'v';
	string[5] = ' ';
	if (order_of_b_men == NORMAL)
	 {
		string[6] = LETTER[b1_man];
		string[7] = b1c + 'a';
		string[8] = b1r + '1';
		string[9] = '/';
		string[10] = LETTER[b2_man];
		string[11] = b2c + 'a';
		string[12] = b2r + '1';
	 }
	else
	 {
		string[6] = LETTER[b2_man];
		string[7] = b2c + 'a';
		string[8] = b2r + '1';
		string[9] = '/';
		string[10] = LETTER[b1_man];
		string[11] = b1c + 'a';
		string[12] = b1r + '1';
	 }
	string[13] = 0;
 }
